/* Apple II+ Simulator */
/* 4.1.93 by Peter */

#include "stdio.h"

/* On Board Ram (48k) of the Apple II+ */
unsigned char ram[0xc000];

/* On Board rom (12k) of the Apple II+ */
unsigned char rom[0x3000];

/* Slots C0SX where S is Slot+8 and X is 0-f */
extern int getC00X();
extern int getC01X();
extern int getC02X();
extern int getC03X();
extern int getC04X();
extern int getC05X();
extern int getC06X();
extern int getC07X();
extern void putC00X();
extern void putC01X();
extern void putC02X();
extern void putC03X();
extern void putC04X();
extern void putC05X();
extern void putC06X();
extern void putC07X();

/* Slots CSXX where S is Slot 1-7 and XX is 00-ff */
extern int getC1XX();
extern int getC2XX();
extern int getC3XX();
extern int getC4XX();
extern int getC5XX();
extern int getC6XX();
extern int getC7XX();
extern void putC1XX();
extern void putC2XX();
extern void putC3XX();
extern void putC4XX();
extern void putC5XX();
extern void putC6XX();
extern void putC7XX();

/* Slots C800-CFFF */
extern int get1C8XX();
extern int get2C8XX();
extern int get3C8XX();
extern int get4C8XX();
extern int get5C8XX();
extern int get6C8XX();
extern int get7C8XX();
extern void put1C8XX();
extern void put2C8XX();
extern void put3C8XX();
extern void put4C8XX();
extern void put5C8XX();
extern void put6C8XX();
extern void put7C8XX();

/* if Rom is inhibited */
extern int get0inhibit();
extern int get1inhibit();
extern int get2inhibit();
extern int get3inhibit();
extern int get4inhibit();
extern int get5inhibit();
extern int get6inhibit();
extern int get7inhibit();
extern void put0inhibit();
extern void put1inhibit();
extern void put2inhibit();
extern void put3inhibit();
extern void put4inhibit();
extern void put5inhibit();
extern void put6inhibit();
extern void put7inhibit();

/* Slot Init */
extern void initslot0();
extern void initslot1();
extern void initslot2();
extern void initslot3();
extern void initslot4();
extern void initslot5();
extern void initslot6();
extern void initslot7();

/* Screen */
extern void initscreen();
extern int getscreen(); /* Keyboard and Paddles */
extern void putscreen();

/* Apple internal IO */
extern int keyboard; /* C000, C010 */
extern int tapeout; /* C020 */
extern int speaker; /* C030 */
extern int strobe; /* C040 */
extern int text; /* C050 */
extern int mixed; /* C052 */
extern int page2; /* C054 */
extern int hires; /* C056 */
extern int an0,an1,an2,an3; /* C058-C05F */
extern int tapein; /* C060 */
extern int button0,button1,button2; /* C061, C062, C063 */
extern int pdl0,pdl1,pdl2,pdl3; /* C064, C065, C066, C067 */
extern int trigger0,trigger1,trigger2,trigger3; /* C070 */

/* Other Flags */
int readinhibit; /* -1 (Rom) or Slot# */
int writeinhibit; /* -1 (Rom) or Slot# */
int slot; /* Slot, der C800 aktiv hat */

void init6502memory()
{
	FILE *f;
	int i;

	/* Read Main Rom */
	f=fopen("f800.rom","r");
	if (!f)
	{
		printf("can't find 'f800.rom'.\n");
		exit(1);
	}
	fread(rom+0x2800,1,0x800,f);
	fclose(f);
	f=fopen("f000.rom","r");
	if (f)
	{
		fread(rom+0x2000,1,0x800,f);
		fclose(f);
	}
	f=fopen("e800.rom","r");
	if (f)
	{
		fread(rom+0x1800,1,0x800,f);
		fclose(f);
	}
	f=fopen("e000.rom","r");
	if (f)
	{
		fread(rom+0x1000,1,0x800,f);
		fclose(f);
	}
	f=fopen("d800.rom","r");
	if (f)
	{
		fread(rom+0x0800,1,0x800,f);
		fclose(f);
	}
	f=fopen("d000.rom","r");
	if (f)
	{
		fread(rom+0x0000,1,0x800,f);
		fclose(f);
	}

	/* Setup buildin IO */
	strobe=0;

	/* Setup other */
	readinhibit= -1;
	writeinhibit= -1;
	slot=0;

	/* Init screen */
	initscreen();

	/* Init slots */
	initslot0();
	initslot1();
	initslot2();
	initslot3();
	initslot4();
	initslot5();
	initslot6();
	initslot7();
}

int get6502memory(adr)
int adr;
{
	if (adr<0xc000)
		return (ram[adr]);
	if (adr>=0xd000)
		switch (readinhibit)
		{
			case 0:
				return (get0inhibit(adr));
			break;

			case 1:
				return (get1inhibit(adr));
			break;

			case 2:
				return (get2inhibit(adr));
			break;

			case 3:
				return (get3inhibit(adr));
			break;

			case 4:
				return (get4inhibit(adr));
			break;

			case 5:
				return (get5inhibit(adr));
			break;

			case 6:
				return (get6inhibit(adr));
			break;

			case 7:
				return (get7inhibit(adr));
			break;

			default:
				return (rom[adr-0xd000]);
			break;
		};
	if (adr==0xcfff)
		slot=0;
	if (adr>=0xc800)
		switch (slot)
		{
			case 0:
				return (0);
			break;

			case 1:
				return (get1C8XX(adr-0xc800));
			break;

			case 2:
				return (get2C8XX(adr-0xc800));
			break;

			case 3:
				return (get3C8XX(adr-0xc800));
			break;

			case 4:
				return (get4C8XX(adr-0xc800));
			break;

			case 5:
				return (get5C8XX(adr-0xc800));
			break;

			case 6:
				return (get6C8XX(adr-0xc800));
			break;

			case 7:
				return (get7C8XX(adr-0xc800));
			break;
		}

	switch ((adr >> 8) & 0x07)
	{
		case 7:
			slot=7;
			return (getC7XX(adr-0xc700));
		break;

		case 6:
			slot=6;
			return (getC6XX(adr-0xc600));
		break;

		case 5:
			slot=5;
			return (getC5XX(adr-0xc500));
		break;

		case 4:
			slot=4;
			return (getC4XX(adr-0xc400));
		break;

		case 3:
			slot=3;
			return (getC3XX(adr-0xc300));
		break;

		case 2:
			slot=2;
			return (getC2XX(adr-0xc200));
		break;

		case 1:
			slot=1;
			return (getC1XX(adr-0xc100));
		break;
	}

	switch ((adr >> 4) & 0x0f)
	{
		case 0x8:
			return(getC00X(adr-0xc080));
		break;

		case 0x9:
			return(getC01X(adr-0xc090));
		break;

		case 0xa:
			return(getC02X(adr-0xc0a0));
		break;

		case 0xb:
			return(getC03X(adr-0xc0b0));
		break;

		case 0xc:
			return(getC04X(adr-0xc0c0));
		break;

		case 0xd:
			return(getC05X(adr-0xc0d0));
		break;

		case 0xe:
			return(getC06X(adr-0xc0e0));
		break;

		case 0xf:
			return(getC07X(adr-0xc0f0));
		break;
	}
	
	return(getscreen(adr));
}

void put6502memory(adr,value)
int adr;
int value;
{
	if (adr<0xc000)
	{
		ram[adr]=value;
		if (((adr>=0x0400)&&(adr<0x0c00))||((adr>=0x2000)&&(adr<0x6000)))
			putscreen(adr,value);
		return;
	}
	if (adr>=0xd000)
		switch (writeinhibit)
		{
			case 0:
				put0inhibit(adr,value);
				return;
			break;

			case 1:
				put1inhibit(adr,value);
				return;
			break;

			case 2:
				put2inhibit(adr,value);
				return;
			break;

			case 3:
				put3inhibit(adr,value);
				return;
			break;

			case 4:
				put4inhibit(adr,value);
				return;
			break;

			case 5:
				put5inhibit(adr,value);
				return;
			break;

			case 6:
				put6inhibit(adr,value);
				return;
			break;

			case 7:
				put7inhibit(adr,value);
				return;
			break;

			default:
			break;
		};
	if (adr==0xcfff)
		slot=0;
	if (adr>=0xc800)
		switch (slot)
		{
			case 0:
				return;
			break;

			case 1:
				put1C8XX(adr-0xc800,value);
				return;
			break;

			case 2:
				put2C8XX(adr-0xc800,value);
				return;
			break;

			case 3:
				put3C8XX(adr-0xc800,value);
				return;
			break;

			case 4:
				put4C8XX(adr-0xc800,value);
				return;
			break;

			case 5:
				put5C8XX(adr-0xc800,value);
				return;
			break;

			case 6:
				put6C8XX(adr-0xc800,value);
				return;
			break;

			case 7:
				put7C8XX(adr-0xc800,value);
				return;
			break;
		}

	switch ((adr >> 8) & 0x07)
	{
		case 7:
			slot=7;
			putC7XX(adr-0xc700,value);
			return;
		break;

		case 6:
			slot=6;
			putC6XX(adr-0xc600,value);
			return;
		break;

		case 5:
			slot=5;
			putC5XX(adr-0xc500,value);
			return;
		break;

		case 4:
			slot=4;
			putC4XX(adr-0xc400,value);
			return;
		break;

		case 3:
			slot=3;
			putC3XX(adr-0xc300,value);
			return;
		break;

		case 2:
			slot=2;
			putC2XX(adr-0xc200,value);
			return;
		break;

		case 1:
			slot=1;
			putC1XX(adr-0xc100,value);
			return;
		break;
	}

	switch ((adr >> 4) & 0x0f)
	{
		case 0x8:
			putC00X(adr-0xc080,value);
			return;
		break;

		case 0x9:
			putC01X(adr-0xc090,value);
			return;
		break;

		case 0xa:
			putC02X(adr-0xc0a0,value);
			return;
		break;

		case 0xb:
			putC03X(adr-0xc0b0,value);
			return;
		break;

		case 0xc:
			putC04X(adr-0xc0c0,value);
			return;
		break;

		case 0xd:
			putC05X(adr-0xc0d0,value);
			return;
		break;

		case 0xe:
			putC06X(adr-0xc0e0,value);
			return;
		break;

		case 0xf:
			putC07X(adr-0xc0f0,value);
			return;
		break;
	}

	value=getscreen(adr);
}
