/* Apple II Simulator */
/* Slotbeschreibung fuer Disk II Controller in Slot S */
/* Peter Koch, 10.2.93 */

#include <stdio.h>
#include <sys/signal.h>
#include <sys/types.h>
#include <sys/stat.h>

/* fuer Slot 6 SIGUSR1, fuer Slot 5 SIGUSR2 benutzen */
#if SLOT==6
#define SIGCHANGE SIGUSR1
#else
#define SIGCHANGE SIGUSR2
#endif

/* External Variables */
extern int readinhibit;
extern int writeinhibit;
extern int slot;

/* Controller Prom */
unsigned char promSLOT[256];

/* Disk-Parameter */
int mediaSLOThaschanged;
int motorSLOT; /* Motor an oder aus */
int driveSLOT; /* Drive 0 oder 1 */
int ddrwSLOT; /* Datenrichtung read oder write */
int diskSLOTbyte;
int protectedSLOT[2];
int trackSLOT[2];
int runbyteSLOT[2];
int sectorSLOT[2];
int volumeSLOT;
int checksumSLOT;
int oldvalueSLOT,exorvalueSLOT;
int wrtrkSLOT,wrsecSLOT;

/* Sektor-Puffer */
unsigned char diskSLOTdata[258];

/* Skewing-Tabelle */

int w_skewingSLOT[16]= { 0,13,11,9,7,5,3,1,14,12,10,8,6,4,2,15 };
int r_skewingSLOT[16]= { 0,7,14,6,13,5,12,4,11,3,10,2,9,1,8,15 };

/* Translation Table */
int translateSLOT[256]= {
0x96, 0x97, 0x9a, 0x9b, 0x9d, 0x9e, 0x9f, 0xa6,
0xa7, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb2, 0xb3,
0xb4, 0xb5, 0xb6, 0xb7, 0xb9, 0xba, 0xbb, 0xbc,
0xbd, 0xbe, 0xbf, 0xcb, 0xcd, 0xce, 0xcf, 0xd3,
0xd6, 0xd7, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde,
0xdf, 0xe5, 0xe6, 0xe7, 0xe9, 0xea, 0xeb, 0xec,
0xed, 0xee, 0xef, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6,
0xf7, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff,
0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x00, 0x01,
0x80, 0x80, 0x02, 0x03, 0x80, 0x04, 0x05, 0x06,
0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x07, 0x08,
0x80, 0x80, 0x80, 0x09, 0x0a, 0x0b, 0x0c, 0x0d,
0x80, 0x80, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13,
0x80, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a,
0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
0x80, 0x80, 0x80, 0x1b, 0x80, 0x1c, 0x1d, 0x1e,
0x80, 0x80, 0x80, 0x1f, 0x80, 0x80, 0x20, 0x21,
0x80, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28,
0x80, 0x80, 0x80, 0x80, 0x80, 0x29, 0x2a, 0x2b,
0x80, 0x2c, 0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32,
0x80, 0x80, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38,
0x80, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f
};

/* File-Pointer */
FILE *fSLOT[2];
char fileSLOTname[2][16];
long fSLOTpos[2];
long fSLOTsize[2];

void diskSLOTchange()
{
	mediaSLOThaschanged=1;
}

void openSLOTfiles()
{
	int i;
	struct stat buf;

	/* Neue Disk ? */
	if (mediaSLOThaschanged)
	{
		mediaSLOThaschanged=0;

		/* Sowohl Drive 1 als auch Drive 2 */
		for (i=0;i<2;i++)
		{
			/* Alte Datei schliessen */
			if (fSLOT[i])
				fclose(fSLOT[i]);

			/* Groesse der neuen Datei bestimmen */
			if (stat(fileSLOTname[i],&buf)<0)
			{
/*
				perror(fileSLOTname[i]);
*/
				fSLOT[i]=NULL;
			}
			else
			{
				fSLOTsize[i]=buf.st_size;

				/* Datei zum lesen und scheiben eroeffnen */
				fSLOT[i]=fopen(fileSLOTname[i],"r+");
				if (fSLOT[i]==NULL)
				{
					/* Na gut, Datei nur zum lesen eroeffnen */
					fSLOT[i]=fopen(fileSLOTname[i],"r");
					if (fSLOT[i]!=NULL)
						protectedSLOT[i]=1;
				}
				else
					protectedSLOT[i]=0;
			}

			/* Auf vernuenftige Werte setzen */
			sectorSLOT[i]=0;
			runbyteSLOT[i]=0;
		}
	}
}

int readSLOTbyte()
{
	int position;
	int value=0x00;

	openSLOTfiles();

	/* Ueberhaupt was da? */
	if (fSLOT[driveSLOT]==NULL)
		return(0xff);

	/* motor an ? */
	if (motorSLOT)
	{
		if (motorSLOT>99)
			return(0x00);
		else
			motorSLOT++;
	}

	/* Was ist zu tun ? */
	switch (runbyteSLOT[driveSLOT])
	{
		case 0:
		case 1: 
		case 2:
		case 3:
		case 4:
		case 5:
		case 20:
		case 21:
		case 22:
		case 23:
		case 24:
			/* Sync */
			value=0xff;
		break;

		case 6:
		case 25:
			/* Header */
			value=0xd5;
		break;

		case 7:
		case 26:
			/* Header */
			value=0xaa;
		break;

		case 8:
			/* Adress-Header */
			value=0x96;
		break;

		case 9:
			/* Volume */
			value=(volumeSLOT >> 1) | 0xaa;
			checksumSLOT=volumeSLOT;
		break;

		case 10:
			/* Volume */
			value=volumeSLOT | 0xaa;
		break;

		case 11:
			/* Track */
			checksumSLOT ^= trackSLOT[driveSLOT]/2;
			value=(trackSLOT[driveSLOT] >> 2) | 0xaa;
		break;

		case 12:
			/* Track */
			value=(trackSLOT[driveSLOT] >> 1) | 0xaa;
		break;

		case 13:
			/* Sector */
			checksumSLOT ^= sectorSLOT[driveSLOT];
			value=(sectorSLOT[driveSLOT] >> 1) | 0xaa;
		break;

		case 14:
			/* Sector */
			value=sectorSLOT[driveSLOT] | 0xaa;
		break;

		case 15:
			/* Checksum */
			value=(checksumSLOT >> 1) | 0xaa;
		break;

		case 16:
			/* Checksum */
			value=checksumSLOT | 0xaa;
		break;

		case 17:
		case 371:
			/* Trailer */
			value=0xde;
		break;

		case 18:
		case 372:
			/* Trailer */
			value=0xaa;
		break;

		case 19:
		case 373:
			/* Trailer */
			value=0xeb;
		break;

		case 27:
			/* Data-Header */
			exorvalueSLOT=0;

			/* Erstes lesen */
			fSLOTpos[driveSLOT]=256*16*(trackSLOT[driveSLOT]/2)+256*r_skewingSLOT[sectorSLOT[driveSLOT]];

			/* Ist ueberhaupt soviel da ? */
			if (fSLOTpos[driveSLOT]+255 > fSLOTsize[driveSLOT])
				return(0xff);

			/* Positionieren */
			fseek(fSLOT[driveSLOT],fSLOTpos[driveSLOT],0);

			/* Sektor lesen */
			fread(diskSLOTdata,1,256,fSLOT[driveSLOT]);
			diskSLOTdata[256]=0;
			diskSLOTdata[257]=0;
			
			value=0xad;
		break;

		case 370:
			/* Checksum */
			value=translateSLOT[exorvalueSLOT & 0x3f];
			sectorSLOT[driveSLOT]++;
			if (sectorSLOT[driveSLOT]==16)
				sectorSLOT[driveSLOT]=0;
		break;

		default:
			position=runbyteSLOT[driveSLOT]-28;
			if (position>=0x56)
			{
				/* 6 Bit */
				position -= 0x56;
				oldvalueSLOT=diskSLOTdata[position];
				oldvalueSLOT=oldvalueSLOT >> 2;
				exorvalueSLOT ^= oldvalueSLOT;
				value=translateSLOT[exorvalueSLOT & 0x3f];
				exorvalueSLOT=oldvalueSLOT;
			}
			else
			{
				/* 3 * 2 Bit */
				oldvalueSLOT=0;
				oldvalueSLOT |= (diskSLOTdata[position] & 0x01) << 1;
				oldvalueSLOT |= (diskSLOTdata[position] & 0x02) >> 1;
				oldvalueSLOT |= (diskSLOTdata[position+0x56] & 0x01) << 3;
				oldvalueSLOT |= (diskSLOTdata[position+0x56] & 0x02) << 1;
				oldvalueSLOT |= (diskSLOTdata[position+0xac] & 0x01) << 5;
				oldvalueSLOT |= (diskSLOTdata[position+0xac] & 0x02) << 3;
				exorvalueSLOT ^= oldvalueSLOT;
				value=translateSLOT[exorvalueSLOT & 0x3f];
				exorvalueSLOT=oldvalueSLOT;
			}
		break;
	}

	/* Weiterschalten */
	runbyteSLOT[driveSLOT]++;
	if (runbyteSLOT[driveSLOT]>373)
		runbyteSLOT[driveSLOT]=0;

	return(value);
}

void writeSLOTbyte(value)
int value;
{
	int position;

	openSLOTfiles();

	/* Ueberhaupt was da? */
	if (fSLOT[driveSLOT]==NULL)
		return;

	if (protectedSLOT[driveSLOT])
		return;

	/* Test fuer doofe */
	if (value<0x96)
		return;

	if (value==0xd5)
		runbyteSLOT[driveSLOT]=6;

	/* Was ist zu tun ? */
	switch (runbyteSLOT[driveSLOT])
	{
		case 0:
		case 1: 
		case 2:
		case 3:
		case 4:
		case 5:
		case 20:
		case 21:
		case 22:
		case 23:
		case 24:
			/* Sync */
		break;

		case 6:
		case 25:
			/* Header */
			if (value==0xff)
				runbyteSLOT[driveSLOT]--;
		break;

		case 7:
		case 26:
			/* Header */
		break;

		case 8:
			/* Adress-Header */
			if (value==0xad)
			{
				exorvalueSLOT=0;
				runbyteSLOT[driveSLOT]=27;
			}
		break;

		case 9:
			/* Volume */
		break;

		case 10:
			/* Volume */
		break;

		case 11:
			/* Track */
			wrtrkSLOT=value << 1;
			wrtrkSLOT &= 0xff;
			wrtrkSLOT |= 0x55;
		break;

		case 12:
			/* Track */
			wrtrkSLOT &= value;
			wrtrkSLOT= wrtrkSLOT << 1;
		break;

		case 13:
			/* Sector */
			wrsecSLOT=value << 1;
			wrsecSLOT &= 0xff;
			wrsecSLOT |= 0x55;
		break;

		case 14:
			/* Sector */
			wrsecSLOT &= value;
			sectorSLOT[driveSLOT] = wrsecSLOT;
		break;

		case 15:
			/* Checksum */
		break;

		case 16:
			/* Checksum */
		break;

		case 17:
		case 371:
			/* Trailer */
		break;

		case 18:
		case 372:
			/* Trailer */
		break;

		case 19:
		case 373:
			/* Trailer */
		break;

		case 27:
			exorvalueSLOT=0;
		break;

		case 370:
			/* Schreibposition */
			fSLOTpos[driveSLOT]=256*16*(trackSLOT[driveSLOT]/2)+256*r_skewingSLOT[sectorSLOT[driveSLOT]];

			/* Ist ueberhaupt soviel da ? */
			if (fSLOTpos[driveSLOT]+255 > fSLOTsize[driveSLOT])
				return;

			/* Positionieren */
			fseek(fSLOT[driveSLOT],fSLOTpos[driveSLOT],0);

			/* Sektor schreiben */
			fwrite(diskSLOTdata,1,256,fSLOT[driveSLOT]);
			fflush(fSLOT[driveSLOT]);
			
			/* Sektor weiterschalten */
			sectorSLOT[driveSLOT]++;
			if (sectorSLOT[driveSLOT]==16)
				sectorSLOT[driveSLOT]=0;
		break;

		default:
			position=runbyteSLOT[driveSLOT]-28;
			value=translateSLOT[value];
			if (position>=0x56)
			{
				/* 6 Bit */
				position -= 0x56;
				value ^= exorvalueSLOT;
				oldvalueSLOT=value;
				diskSLOTdata[position] |= (value << 2) & 0xfc ;
				exorvalueSLOT=oldvalueSLOT;
			}
			else
			{
				/* 3 * 2 Bit */
				value ^= exorvalueSLOT;
				oldvalueSLOT=value;
				diskSLOTdata[position] = (value & 0x01) << 1;
				diskSLOTdata[position] |= (value & 0x02) >> 1;
				diskSLOTdata[position+0x56] = (value & 0x04) >> 1;
				diskSLOTdata[position+0x56] |= (value & 0x08) >> 3;
				diskSLOTdata[position+0xac] = (value & 0x10) >> 3;
				diskSLOTdata[position+0xac] |= (value & 0x20) >> 5;
				exorvalueSLOT=oldvalueSLOT;
			}
		break;
	}

	/* Weiterschalten */
	runbyteSLOT[driveSLOT]++;
	if (runbyteSLOT[driveSLOT]>373)
		runbyteSLOT[driveSLOT]=0;

}

int getC0SLOTX(adr)
int adr;
{
	int phase;

	switch (adr & 0x000f)
	{
		case 0x08:
			motorSLOT=1;
		break;

		case 0x09:
			motorSLOT=0;
		break;

		case 0x0a:
			driveSLOT=0;
		break;

		case 0x0b:
			driveSLOT=1;
		break;

		case 0x0c:
			if (ddrwSLOT==0)
				return(readSLOTbyte());
			else
				writeSLOTbyte(diskSLOTbyte);
		break;

		case 0x0d:
			if (protectedSLOT[driveSLOT])
				return(0x80);
		break;

		case 0x0e:
			ddrwSLOT=0;
			if (protectedSLOT[driveSLOT])
				return(0x80);
		break;

		case 0x0f:
			ddrwSLOT=1;
		break;

		default:
			/* Schrittmotor */
			phase=(adr & 0x06) >> 1;
			if (adr & 0x01)
			{
				phase -= (trackSLOT[driveSLOT] % 4);
				if (phase<0) phase += 4;
				if (phase==1)
					trackSLOT[driveSLOT]++;
				if (phase==3)
					trackSLOT[driveSLOT]--;

				if (trackSLOT[driveSLOT]<0)
					trackSLOT[driveSLOT]=0;
			}
		break;
	}
	return(0);
}

void putC0SLOTX(adr,value)
int adr;
int value;
{
	if ((adr & 0x000f) == 0x000d)
		diskSLOTbyte=value;
	else
		value=getC0SLOTX(adr);
}

int getCSLOTXX(adr)
int adr;
{
	return(promSLOT[adr & 0x00ff]);
}

void putCSLOTXX(adr,value)
int adr;
int value;
{
	/* Unused */
}

int getSLOTC8XX(adr)
int adr;
{
	/* Unused */
	return(0);
}

void putSLOTC8XX(adr,value)
int adr;
int value;
{
	/* Unused */
}

int getSLOTinhibit(adr)
int adr;
{
	/* Unused */
	return(0);
}

void putSLOTinhibit(adr,value)
int adr;
int value;
{
	/* Unused */
}

void initslotSLOT()
{
	int i;
	FILE *g;

	/* Read prom */
	g=fopen("controller.rom","r");
	if (g==0)
	{
		fprintf(stderr,"can't find 'controller.rom'.\n");
		exit(1);
	}
	fread(promSLOT,1,0x0100,g);
	fclose(g);

	/* Install signal-Routine */
	signal(SIGCHANGE,diskSLOTchange);

	/* Read Disk Images for the first time */
	mediaSLOThaschanged=1;
	fSLOT[0]=NULL;
	fSLOT[1]=NULL;

	trackSLOT[0]=42;
	trackSLOT[1]=42;

	/* set filenames */
	sprintf(fileSLOTname[0],"s%dd1",SLOT);
	sprintf(fileSLOTname[1],"s%dd2",SLOT);

	/* set variables */
	motorSLOT=1;
	driveSLOT=0;
	ddrwSLOT=0;
	volumeSLOT=254;
}
