/* 6502 Simulator */
/* 30.12.92 by Peter */

/* Debugging: if undefined, no code is generated. */
/* #define DISASM if (cpustate) if (getdisasm) */
#define DISASM if (0)

/* Interface for memory-mapped I/O */
extern int get6502memory();
extern void put6502memory();

/* Register: uchar=8bit ushort=16bit assumed!!! */
unsigned char accu,xreg,yreg,flags,stack;
unsigned short pcnt;

/* internal registers */
unsigned char opcode;
unsigned char cpustate;

/* Macros for cpustate */
#define setreset cpustate |= 0x80
#define setirq cpustate |= 0x40
#define setnmi cpustate |= 0x20
#define setdisasm cpustate |= 0x10
#define setregshow cpustate |= 0x08
#define setunused1 cpustate |= 0x04
#define setunused2 cpustate |= 0x02
#define setunused3 cpustate |= 0x01

#define clrreset cpustate &= ~0x80
#define clrirq cpustate &= ~0x40
#define clrnmi cpustate &= ~0x20
#define clrdisasm cpustate &= ~0x10
#define clrregshow cpustate &= ~0x08
#define clrunused1 cpustate &= ~0x04
#define clrunused2 cpustate &= ~0x02
#define clrunused3 cpustate &= ~0x01

#define getreset ((cpustate & 0x80)>>7)
#define getirq ((cpustate & 0x40)>>6)
#define getnmi ((cpustate & 0x20)>>5)
#define getdisasm ((cpustate & 0x10)>>4)
#define getregshow ((cpustate & 0x08)>>3)
#define getunused1 ((cpustate & 0x04)>>2)
#define getunused2 ((cpustate & 0x02)>>1)
#define getunused3 (cpustate & 0x01)

/* Macros for convenience */
#define A accu
#define X xreg
#define Y yreg
#define P flags
#define S stack
#define PC pcnt

/* Macros for processor flags */
#define set6502n P |= 0x80
#define set6502v P |= 0x40
#define set6502x P |= 0x20
#define set6502b P |= 0x10
#define set6502d P |= 0x08
#define set6502i P |= 0x04
#define set6502z P |= 0x02
#define set6502c P |= 0x01

#define clr6502n P &= ~0x80
#define clr6502v P &= ~0x40
#define clr6502x P &= ~0x20
#define clr6502b P &= ~0x10
#define clr6502d P &= ~0x08
#define clr6502i P &= ~0x04
#define clr6502z P &= ~0x02
#define clr6502c P &= ~0x01

#define get6502n ((P & 0x80)>>7)
#define get6502v ((P & 0x40)>>6)
#define get6502x ((P & 0x20)>>5)
#define get6502b ((P & 0x10)>>4)
#define get6502d ((P & 0x08)>>3)
#define get6502i ((P & 0x04)>>2)
#define get6502z ((P & 0x02)>>1)
#define get6502c (P & 0x01)

#define set6502nz(reg) { if (reg) clr6502z; else set6502z; if (reg & 0x80) set6502n; else clr6502n; }

/* geneal Macro for handling 16bit adresses */
#define get6502adress(adr) ((get6502memory(adr+1)<<8) | (get6502memory(adr)))

/* general Macros for Stack */
#define push6502(value) put6502memory(0x100+S--,value)
#define pull6502 get6502memory(++S+0x100)

#define push6502flags push6502(P)
#define push6502accu push6502(A)
#define push6502pc push6502(PC/0x100);push6502(PC%0x100)

/* Init CPU */
void init6502()
{
	setreset;
}

/* Reset cpu */
void reset6502()
{
	setreset;
}

/* Non maskable interrupt */
void nmi6502()
{
	setnmi;
}

/* interrupt */
void irq6502()
{
	setirq;
}

void exec6502()
{
	unsigned char operation,adrmode,value,saveflags;
	unsigned short savepc;
	int sum;

	if (cpustate)
	{
		if (getregshow)
		{
			printf("A: %2x, ",A);
			printf("X: %2x, ",X);
			printf("Y: %2x, ",Y);
			printf("S: %2x, ",S);
			printf("P: %2x, ",P);
			printf("PC: %4x\n",PC);
		}

		if (getreset)
		{
			DISASM printf("reset\n");
			A=X=Y=P=0;
			S=0xff;
			PC=get6502adress(0xfffc);
			clrreset;
		}

		if (getnmi)
		{
			DISASM printf("nmi\n");
			push6502pc;
			push6502flags;
			set6502i;
			PC=get6502adress(0xfffa);
			clrnmi;
		}

		if (getirq)
			if (get6502i==0)
			{
				DISASM printf("irq\n");
				push6502pc;
				push6502flags;
				set6502i;
				PC=get6502adress(0xfffe);
				clrirq;
			}
	}

	opcode=get6502memory(PC++);
	
	DISASM
		printf("%2x ",opcode);

	switch (opcode)
	{
		/* opcodes not matching any adrmode */
		case 0x0a: /* ASL A */
			if (A & 0x80) set6502c; else clr6502c;
			A = A << 1;
			set6502nz(A);
			DISASM
				printf("        Accu    ASL\n");
		break;

		case 0x00: /* BRK */
			PC++;
			push6502pc;
			push6502flags;
			set6502i;
			set6502b;
			PC=get6502adress(0xfffe);
			DISASM
				printf("        Implied BRK\n");
		break;

		case 0x24: /* BIT ZP */
			savepc=get6502memory(PC++);
			value=get6502memory(savepc);
			if (value & 0x80) set6502n; else clr6502n;
			if (value & 0x40) set6502v; else clr6502v;
			if (value & A) clr6502z; else set6502z;
			DISASM
				printf("%2x      ZP      BIT\n",savepc);
		break;

		case 0x2c: /* BIT Abs */
			savepc=get6502adress(PC);
			PC++;PC++;
			value=get6502memory(savepc);
			if (value & 0x80) set6502n; else clr6502n;
			if (value & 0x40) set6502v; else clr6502v;
			if (value & A) clr6502z; else set6502z;
			DISASM
				printf("%4x    Abs     BIT\n",savepc);
		break;

		case 0x18: /* CLC */
			clr6502c;
			DISASM
				printf("        Implied CLC\n");
		break;

		case 0xd8: /* CLD */
			clr6502d;
			DISASM
				printf("        Implied CLD\n");
		break;

		case 0x58: /* CLI */
			clr6502i;
			DISASM
				printf("        Implied CLI\n");
		break;

		case 0xb8: /* CLV */
			clr6502v;
			DISASM
				printf("        Implied CLV\n");
		break;
	
		case 0xca: /* DEX */
			X--;
			set6502nz(X);
			DISASM
				printf("        Implied DEX\n");
		break;

		case 0x88: /* DEY */
			Y--;
			set6502nz(Y);
			DISASM
				printf("        Implied DEY\n");
		break;

		case 0xe8: /* INX */
			X++;
			set6502nz(X);
			DISASM
				printf("        Implied INX\n");
		break;

		case 0xc8: /* INY */
			Y++;
			set6502nz(Y);
			DISASM
				printf("        Implied INY\n");
		break;

		case 0x4c: /* JMP xxxx */
			PC=get6502adress(PC);
			DISASM
				printf("%4x    Abs     JMP\n",PC);
		break;

		case 0x6c: /* JMP (xxxx) */
			savepc=get6502adress(PC);
			PC=get6502adress(savepc);
			DISASM
				printf("%4x    Ind     JMP\n",savepc);
		break;

		case 0x20: /* JSR xxxx */
			PC++;
			push6502pc;
			PC--;
			PC=get6502adress(PC);
			DISASM
				printf("%4x    Abs     JSR\n",PC);
		break;

		case 0xb6: /* LDX ZP,Y */
			savepc=get6502memory(PC++)+Y;
			X=get6502memory(savepc & 0xff);
			set6502nz(X);
			DISASM
				printf("%2x      ZP,Y    LDX\n",savepc-Y);
		break;

		case 0xbe: /* LDX Abs,Y */
			savepc=get6502adress(PC)+Y;
			PC++;PC++;
			X=get6502memory(savepc);
			set6502nz(X);
			DISASM
				printf("%4x    Abs,Y   LDX\n",savepc-Y);
		break;

		case 0x4a: /* LSR A */
			if (A & 0x01 ) set6502c; else clr6502c;
			A = A >> 1;
			set6502nz(A);
			DISASM
				printf("        Accu    LSR\n");
		break;

		case 0xea: /* NOP */
			DISASM
				printf("        Implied NOP\n");
		break;

		case 0x48: /* PHA */
			push6502accu;
			DISASM
				printf("        Implied PHA\n");
		break;

		case 0x08: /* PHP */
			push6502flags;
			DISASM
				printf("        Implied PHP\n");
		break;

		case 0x68: /* PLA */
			A=pull6502;
			set6502nz(A);
			DISASM
				printf("        Implied PLA\n");
		break;

		case 0x28: /* PLP */
			P=pull6502;
			set6502x;
			DISASM
				printf("        Implied PLP\n");
		break;

		case 0x2a: /* ROL A */
			saveflags=get6502c;
			if (A & 0x80) set6502c; else clr6502c;
			A = A << 1;
			if (saveflags) A |= 0x01;
			set6502nz(A);
			DISASM
				printf("        Accu    ROL\n");
		break;

		case 0x6a: /* ROR A */
			saveflags=get6502c;
			if (A & 0x01 ) set6502c; else clr6502c;
			A = A >> 1;
			if (saveflags) A |= 0x80;
			set6502nz(A);
			DISASM
				printf("        Accu    ROR\n");
		break;

		case 0x40: /* RTI */
			P=pull6502;
			set6502x;
			PC=pull6502;
			PC |= pull6502<<8;
			DISASM
				printf("        Implied RTI\n");
		break;

		case 0x60: /* RTS */
			PC=pull6502;
			PC |= pull6502<<8;
			PC++;
			DISASM
				printf("        Implied RTS\n");
		break;

		case 0x38: /* SEC */
			set6502c;
			DISASM
				printf("       Implied SEC\n");
		break;

		case 0xf8: /* SED */
			set6502d;
			DISASM
				printf("       Implied SED\n");
		break;

		case 0x78: /* SEI */
			set6502i;
			DISASM
				printf("        Implied SEI\n");
		break;

		case 0x96: /* STX ZP,Y */
			savepc=get6502memory(PC++)+Y;
			put6502memory(savepc & 0xff,X);
			DISASM
				printf("%2x      ZP,Y    STX\n",savepc-Y);
		break;

		case 0xaa: /* TAX */
			X=A;
			set6502nz(X);
			DISASM
				printf("        Implied TAX\n");
		break;

		case 0xa8: /* TAY */
			Y=A;
			set6502nz(Y);
			DISASM
				printf("        Implied TAY\n");
		break;

		case 0xba: /* TSX */
			X=S;
			set6502nz(X);
			DISASM
				printf("        Implied TSX\n");
		break;

		case 0x8a: /* TXA */
			A=X;
			set6502nz(A);
			DISASM
				printf("        Implied TXA\n");
		break;

		case 0x9a: /* TXS */
			S=X;
			DISASM
				printf("        Implied TXS\n");
		break;

		case 0x98: /* TYA */
			A=Y;
			set6502nz(A);
			DISASM
				printf("        Implied TYA\n");
		break;

		default:
			adrmode=(opcode & 0x1f);
			operation=(opcode & 0xe1);
			switch (adrmode)
			{
				case 0x1:
					/* (ZP,X) */
					value=get6502memory(PC++)+X;
					savepc=get6502adress(value);
					DISASM
						printf("%2x       (ZP,X)  ",value);
				break;

				case 0x0:
				case 0x2:
				case 0x9:
					/* Immediate */
					savepc=PC++;
					DISASM
						printf("%2x      Imm.    ",get6502memory(savepc));
				break;

				case 0x4:
				case 0x5:
				case 0x6:
					/* ZP */
					savepc=get6502memory(PC++);
					DISASM
						printf("%2x      ZP      ",savepc);
				break;

				case 0xc:
				case 0xd:
				case 0xe:
					/* Absolute */
					savepc=get6502adress(PC);
					PC++;PC++;
					DISASM
						printf("%4x    Abs     ",savepc);
				break;

				case 0x10:
					/* Branch relative */
					savepc=get6502memory(PC++);
					DISASM
						printf("%2x      Rel.    ",savepc);
					if (savepc & 0x80) savepc -= 0x100;
				break;

				case 0x11:
					/* (ZP),Y */
					value=get6502memory(PC++);
					savepc=get6502adress(value)+Y;
					DISASM
						printf("%2x      (ZP),Y  ",value);
				break;

				case 0x14:
				case 0x15:
				case 0x16:
					/* ZP,X */
					savepc=get6502memory(PC++);
					DISASM
						printf("%2x      ZP,X    ",savepc);
					savepc=(savepc+X) & 0xff;
				break;

				case 0x19:
					/* Absolute,Y */
					savepc=get6502adress(PC)+Y;
					PC++;PC++;
					DISASM
						printf("%4x    Abs,Y   ",savepc-Y);
				break;

				case 0x1c:
				case 0x1d:
				case 0x1e:
					/* Absolute,X */
					savepc=get6502adress(PC)+X;
					PC++;PC++;
					DISASM
						printf("%4x    Abs,X   ",savepc-X);
				break;

				case 0x3:
				case 0x7:
				case 0x8:
				case 0xa:
				case 0xb:
				case 0xf:
				case 0x12:
				case 0x13:
				case 0x17:
				case 0x18:
				case 0x1a:
				case 0x1b:
				case 0x1f:
					/* No Op */
					DISASM
						printf("        Illegal NOP\n");
					return;
				break;
			}
			if (adrmode==0x10)
			{
				/* Branch */
				switch ((operation)>>5)
				{
					case 0: /* BPL */
						if (!get6502n) PC += savepc;
						DISASM
							printf("BPL\n");
					break;

					case 1: /* BMI */
						if (get6502n) PC += savepc;
						DISASM
							printf("BMI\n");
					break;

					case 2: /* BVC */
						if (!get6502v) PC += savepc;
						DISASM
							printf("BVC\n");
					break;

					case 3: /* BVS */
						if (get6502v) PC += savepc;
						DISASM
							printf("BVS\n");
					break;

					case 4: /* BCC */
						if (!get6502c) PC += savepc;
						DISASM
							printf("BCC\n");
					break;

					case 5: /* BCS */
						if (get6502c) PC += savepc;
						DISASM
							printf("BCS\n");
					break;

					case 6: /* BNE */
						if (!get6502z) PC += savepc;
						DISASM
							printf("BNE\n");
					break;

					case 7: /* BEQ */
						if (get6502z) PC += savepc;
						DISASM
							printf("BEQ\n");
					break;
				}
			}
			else
			{
				/* Normal Operation */
				switch (operation)
				{
					case 0x00: /* ASL */
						value=get6502memory(savepc);
						if (value & 0x80) set6502c; else clr6502c;
						value = value << 1;
						put6502memory(savepc,value);
						set6502nz(value);
						DISASM
							printf("ASL\n");
					break;

					case 0x01: /* ORA */
						value=get6502memory(savepc);
						A |= value;
						set6502nz(A);
						DISASM
							printf("ORA\n");
					break;

					case 0x20: /* ROL */
						value=get6502memory(savepc);
						saveflags=get6502c;
						if (value & 0x80) set6502c; else clr6502c;
						value = value << 1;
						if (saveflags) value |= 0x01;
						put6502memory(savepc,value);
						set6502nz(value);
						DISASM
							printf("ROL\n");
					break;
		
					case 0x21: /* AND */
						value=get6502memory(savepc);
						A &= value;
						set6502nz(A);
						DISASM
							printf("AND\n");
					break;

					case 0x40: /* LSR */
						value=get6502memory(savepc);
						if (value & 0x01 ) set6502c; else clr6502c;
						value = value >> 1;
						put6502memory(savepc,value);
						set6502nz(value);
						DISASM
							printf("LSR\n");
					break;
		
					case 0x41: /* EOR */
						value=get6502memory(savepc);
						A ^= value;
						set6502nz(A);
						DISASM
							printf("EOR\n");
					break;

					case 0x60: /* ROR */
						value=get6502memory(savepc);
						saveflags=get6502c;
						if (value & 0x01 ) set6502c; else clr6502c;
						value = value >> 1;
						if (saveflags) value |= 0x80;
						put6502memory(savepc,value);
						set6502nz(value);
						DISASM
							printf("ROR\n");
					break;
		
					case 0x61: /* ADC (ohne Dezimal) */
						value=get6502memory(savepc);
						saveflags=get6502c;
						sum = ((char)A) + ((char)value) + saveflags;
						if ((sum>0x7f)||(sum<-0x80)) set6502v; else clr6502v;
						sum = A + value + saveflags;
						if (sum>0xff) set6502c; else clr6502c;
						A = sum;
						if (get6502d)
						{
							clr6502c;
							if ((A & 0x0f)>0x09)
								A += 0x06;
							if ((A & 0xf0)>0x90)
							{
								A += 0x60;
								set6502c;
							}
						}
						else
						{
							set6502nz(A);
						}
						DISASM
							printf("ADC\n");
					break;

					case 0x80: /* STX/STY */
						if (opcode & 0x02)
							put6502memory(savepc,X);
						else
							put6502memory(savepc,Y);
						DISASM
							if (opcode & 0x02)
								printf("STX\n");
							else
								printf("STY\n");
					break;
		
					case 0x81: /* STA */
						put6502memory(savepc,A);
						DISASM
							printf("STA\n");
					break;

					case 0xa0: /* LDX/LDY */
						value=get6502memory(savepc);
						if (opcode & 0x02)
							X=value;
						else
							Y=value;
						set6502nz(value);
						DISASM
							if (opcode & 0x02)
								printf("LDX\n");
							else
								printf("LDY\n");
					break;
		
					case 0xa1: /* LDA */
						value=get6502memory(savepc);
						A=value;
						set6502nz(A);
						DISASM
							printf("LDA\n");
					break;

					case 0xc0: /* DEC/CPY */
						if (opcode & 0x02)
						{
							value=get6502memory(savepc);
							put6502memory(savepc,--value);
							set6502nz(value);
							DISASM
								printf("DEC\n");
						}
						else
						{
							value=get6502memory(savepc);
							if (Y+0x100-value>0xff) set6502c; else clr6502c;
							value=Y+0x100-value;
							set6502nz(value);
							DISASM
								printf("CPY\n");
						}
					break;
		
					case 0xc1: /* CMP */
						value=get6502memory(savepc);
						if (A+0x100-value>0xff) set6502c; else clr6502c;
						value=A+0x100-value;
						set6502nz(value);
						DISASM
							printf("CMP\n");
					break;

					case 0xe0: /* INC/CPX */
						if (opcode & 0x02)
						{
							value=get6502memory(savepc);
							put6502memory(savepc,++value);
							set6502nz(value);
							DISASM
								printf("INC\n");
						}
						else
						{
							value=get6502memory(savepc);
							if (X+0x100-value>0xff) set6502c; else clr6502c;
							value=X+0x100-value;
							set6502nz(value);
							DISASM
								printf("CPX\n");
						}
					break;
		
					case 0xe1: /* SBC (ohne Dezimal)*/
						value=get6502memory(savepc);
						saveflags=!get6502c;
						sum = ((char)A) - ((char)value) - saveflags;
						if ((sum>0x7f)||(sum<-0x80)) set6502v; else clr6502v;
						sum = A - value - saveflags;
						if (sum>=0) set6502c; else clr6502c;
						A = sum;
						if (get6502d)
						{
							A -= 0x66;
							clr6502c;
							if ((A & 0x0f)>0x09)
								A += 0x06;
							if ((A & 0xf0)>0x90)
							{
								A += 0x60;
								set6502c;
							}
						}
						else
						{
							set6502nz(A);
						}
						DISASM
							printf("SBC\n");
					break;
				}
			}
		break;
	}
}
