/* Apple II Color-Screen-Emulator */
/* Peter Koch, 5.1.93 */

#include <X11/Xlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <signal.h>

/* Character Rom */
unsigned char crom[0x0800];

/* Apple Ram */
extern unsigned char ram[0xc000];

/* debug flags */
extern int solid,beep,monochrome,flush,keep;

/* Groesse eines Pixels */
#define PIXSIZE 2

/* Grau-Bitmaps fuer Lores (PIXSIZE=2) */
#if (PIXSIZE==2)
#define grey1_width 16
#define grey1_height 8
static char grey1_bits[] = {
0x55, 0x55, 0xaa, 0xaa, 0x55, 0x55, 0xaa, 0xaa,
0x55, 0x55, 0xaa, 0xaa, 0x55, 0x55, 0xaa, 0xaa
};

#define grey2_width 16
#define grey2_height 8
static char grey2_bits[] = {
0x22, 0x22, 0x88, 0x88, 0x22, 0x22, 0x88, 0x88,
0x22, 0x22, 0x88, 0x88, 0x22, 0x22, 0x88, 0x88
};

#define grey3_width 16
#define grey3_height 8
static char grey3_bits[] = {
0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55,
0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55
};

#define grey4_width 16
#define grey4_height 8
static char grey4_bits[] = {
0x77, 0x77, 0xdd, 0xdd, 0x77, 0x77, 0xdd, 0xdd,
0x77, 0x77, 0xdd, 0xdd, 0x77, 0x77, 0xdd, 0xdd
};
#endif

/* Grau-Bitmaps fuer Lores (PIXSIZE=1) */
#if (PIXSIZE==1)
#define grey1_width 8
#define grey1_height 4
static char grey1_bits[] = { 0x55, 0xaa, 0x55, 0xaa };

#define grey2_width 8
#define grey2_height 4
static char grey2_bits[] = { 0x22, 0x88, 0x22, 0x88 };

#define grey3_width 8
#define grey3_height 4
static char grey3_bits[] = { 0x55, 0x55, 0x55, 0x55 };

#define grey4_width 8
#define grey4_height 4
static char grey4_bits[] = { 0x77, 0xdd, 0x77, 0xdd };
#endif

/* Bascalc-Table */
int bascalc[24]=
{
0x000, 0x080, 0x100, 0x180, 0x200, 0x280, 0x300, 0x380,
0x028, 0x0a8, 0x128, 0x1a8, 0x228, 0x2a8, 0x328, 0x3a8,
0x050, 0x0d0, 0x150, 0x1d0, 0x250, 0x2d0, 0x350, 0x3d0
};

/* Apple I/O */
int keyboard; /* C000, C010 */
int tapeout; /* C020 */
int speaker; /* C030 */
int strobe; /* C040 */
int text; /* C050 */
int mixed; /* C052 */
int page2; /* C054 */
int hires; /* C056 */
int an0,an1,an2,an3; /* C058-C05F */
int tapein; /* C060 */
int button0,button1,button2; /* C061, C062, C063 */
int pdl0,pdl1,pdl2,pdl3; /* C064, C065, C066, C067 */
int trigger0,trigger1,trigger2,trigger3; /* C070 */

/* Various X-Variables */
Display *display;
Window rw,w;
GC gc,rgc;
XSetWindowAttributes attrib;
int screen;
int depth;
Colormap colormap;
XColor color,exact;
char *displayname;
int black,white,green,purple,orange,blue;
int lorescolor[16];
Pixmap text1,text2,lores1,lores2,hires1,hires2;
Pixmap charset,hiresset,loresset;
Pixmap greymap,helpmap;

#define drawchar(pm,row,col,value) XCopyArea(display,charset,*pm,gc,PIXSIZE*8*value,0,PIXSIZE*7,PIXSIZE*8,PIXSIZE*7*col,PIXSIZE*8*row);
#define drawblock(pm,row,col,value) XCopyArea(display,loresset,*pm,gc,PIXSIZE*(8*value+(col & 1)),0,PIXSIZE*7,PIXSIZE*8,PIXSIZE*7*col,PIXSIZE*8*row);
#define drawdots(pm,row,col,crow,value) XCopyArea(display,hiresset,*pm,gc,PIXSIZE*8*value,PIXSIZE*(col & 1),PIXSIZE*7,PIXSIZE,PIXSIZE*7*col,PIXSIZE*(8*row+crow));

void redrawscreen()
{
	int msize;
	int row,col,crow;
	Pixmap *pm;
	int adr;

	if (!keep)
	{
		/* Discard unused Pixmaps */
		if (((!text)&&(!mixed))||(text&&(page2)))
		{
			if (text1)
				XFreePixmap(display,text1);
			text1=NULL;
		}
	
		if (((!text)&&(!mixed))||(text&&(!page2)))
		{
			if (text2)
				XFreePixmap(display,text2);
			text2=NULL;
		}
	
		if (hires || text)
		{
			if (lores1)
				XFreePixmap(display,lores1);
			lores1=NULL;
			if (lores2)
				XFreePixmap(display,lores2);
			lores2=NULL;
		}
		
		if ((!hires) || text)
		{
			if (hires1)
				XFreePixmap(display,hires1);
			hires1=NULL;
			if (hires2)
				XFreePixmap(display,hires2);
			hires2=NULL;
		}
	}

	/* Draw Text */
	if (mixed) msize=160;
	if (text) msize=0;

	if (text || mixed)
	{
		if (page2)
		{
			pm= &text2;
			adr=0x0800;
		}
		else
		{
			pm= &text1;
			adr=0x0400;
		}

		if (*pm==NULL)
		{
			*pm=XCreatePixmap(display,w,PIXSIZE*280,PIXSIZE*192,depth);
			XFillRectangle(display,*pm,rgc,0,0,PIXSIZE*280,PIXSIZE*192);
			for (row=0;row<24;row++)
				for (col=0;col<40;col++)
					drawchar(pm,row,col,ram[adr+bascalc[row]+col]);
		}
		XCopyArea(display,*pm,w,gc,0,PIXSIZE*msize,PIXSIZE*280,PIXSIZE*(192-msize),0,PIXSIZE*msize);
	}

	/* Draw Graphic */	
	msize=192;
	if (mixed) msize=160;

	if (text==0)
	{
		if (hires)
		{
			if (page2)
			{
				pm= &hires2;
				adr=0x4000;
			}
			else
			{
				pm= &hires1;
				adr=0x2000;
			}
			if (*pm==NULL)
			{
				*pm=XCreatePixmap(display,w,PIXSIZE*280,PIXSIZE*192,depth);
				XFillRectangle(display,*pm,rgc,0,0,PIXSIZE*280,PIXSIZE*192);
				for (row=0;row<24;row++)
					for (col=0;col<40;col++)
						for (crow=0;crow<8;crow++)
							drawdots(pm,row,col,crow,ram[adr+1024*crow+bascalc[row]+col]);
			}
		}
		else
		{
			if (page2)
			{
				pm= &lores2;
				adr=0x0800;
			}
			else
			{
				pm= &lores1;
				adr=0x0400;
			}
			if (*pm==NULL)
			{
				*pm=XCreatePixmap(display,w,PIXSIZE*280,PIXSIZE*192,depth);
				XFillRectangle(display,*pm,rgc,0,0,PIXSIZE*280,PIXSIZE*192);
				for (row=0;row<24;row++)
					for (col=0;col<40;col++)
						drawblock(pm,row,col,ram[adr+bascalc[row]+col]);
			}
		}
		
		XCopyArea(display,*pm,w,gc,0,0,PIXSIZE*280,PIXSIZE*msize,0,0);
	}

	XFlush(display);
}

int getscreen(adr)
int adr;
{
	XEvent evnt;
	XEvent *ev;
	int value,count;
	char buf[256];
	char redraw;

	value=0;
	redraw=0;
	XFlush(display);
	/* Parsing X-Events */
	ev= &evnt;
	while (XEventsQueued(display,QueuedAfterReading))
	{
		XNextEvent(display,ev);
		switch (ev->type)
		{
			case ButtonPress:
				if (((XButtonEvent *)ev)->button == 1)
					button0=0x80;
				if (((XButtonEvent *)ev)->button == 2)
					button1=0x80;
				if (((XButtonEvent *)ev)->button == 3)
					button2=0x80;
			break;

			case ButtonRelease:
				if (((XButtonEvent *)ev)->button == 1)
					button0=0;
				if (((XButtonEvent *)ev)->button == 2)
					button1=0;
				if (((XButtonEvent *)ev)->button == 3)
					button2=0;
			break;

			case MotionNotify:
				pdl0=(((XMotionEvent *)ev)->x)/PIXSIZE;
				pdl1=(((XMotionEvent *)ev)->y)/PIXSIZE;
				if (pdl0>255) pdl0=255;
			break;

			case KeyPress:
				count=XLookupString(ev,buf,256,NULL,NULL);
				if (count)
					keyboard=buf[0] | 0x80;
			break;

			case Expose:
				redraw=1;
			break;
		}
	}

	/* Ring bell and look for soft switches */
	switch (adr & 0xf0)
	{
		case 0x00: /* Keyboard */
			value=keyboard;
			if (value & 0x40)
				value ^= 0x20;
		break;

		case 0x10: /* Keyboard Strobe */
			keyboard &= 0x7f;
		break;

		case 0x20: /* Tapeout */
			tapeout= !tapeout;
		break;

		case 0x30: /* Speaker */
			if (beep)
			{
				speaker++;
				if (speaker>191)
				{
					XBell(display,50);
					speaker=0;
				}
			}
		break;

		case 0x40: /* I/O-Strobe */
			strobe=0;
			strobe=1;
		break;

		case 0x50: /* Softswitches */
			switch (adr & 0x000f)
			{
				case 0:
					if (text)
						redraw=1;
					text=0;
				break;

				case 1:
					if (!text)
						redraw=1;
					text=1;
				break;

				case 2:
					if ((mixed)&&(!text))
						redraw=1;
					mixed=0;
				break;

				case 3:
					if ((!mixed)&&(!text))
						redraw=1;
					mixed=1;
				break;

				case 4:
					if (page2)
						redraw=1;
					page2=0;
				break;

				case 5:
					if (!page2)
						redraw=1;
					page2=1;
				break;

				case 6:
					if ((hires)&&(!text))
						redraw=1;
					hires=0;
				break;

				case 7:
					if ((!hires)&&(!text))
						redraw=1;
					hires=1;
				break;

				case 8:
					an0=0;
				break;

				case 9:
					an0=1;
				break;

				case 10:
					an1=0;
				break;

				case 11:
					an1=1;
				break;

				case 12:
					an2=0;
				break;

				case 13:
					an2=1;
				break;

				case 14:
					an3=0;
				break;

				case 15:
					an3=1;
				break;
			}
		break;

		case 0x60: /* Paddles und Buttons */
			switch (adr & 0x0007)
			{
				case 0:
					value=tapein;
				break;

				case 1:
					value=button0;
				break;

				case 2:
					value=button1;
				break;

				case 3:
					value=button2;
				break;

				case 4:
					if (trigger0<pdl0)
						value=0xff;
					else
						value=0x00;
					if (trigger0<0xff)
						trigger0++;
				break;

				case 5:
					if (trigger1<pdl1)
						value=0xff;
					else
						value=0x00;
					if (trigger1<0xff)
						trigger1++;
				break;

				case 6:
					if (trigger2<pdl2)
						value=0xff;
					else
						value=0x00;
					if (trigger2<0xff)
						trigger2++;
				break;

				case 7:
					if (trigger3<pdl3)
						value=0xff;
					else
						value=0x00;
					if (trigger3<0xff)
						trigger3++;
				break;
			}
		break;

		case 0x70: /* Trigger */
			trigger0=0;
			trigger1=0;
			trigger2=0;
			trigger3=0;
		break;
	}
	
	/* redraw screen if necessary */
	if (redraw)
		redrawscreen();

	return(value);
}

void putscreen(adr,value)
int adr;
int value;
{
	int row,col,crow;

	/* Calculate visible position */
	row=(adr & 0x3ff)/128;
	col=(adr & 0x3ff)-bascalc[row];
	row=row+8*(col/40);
	col=(adr & 0x3ff)-bascalc[row];

	/* Non-visible cells */
	if (row>23)
		return;

	/* text or lores */
	if ((adr>=0x400)&&(adr<0x800))
	{
		if (text1)
			drawchar(&text1,row,col,value);
		if (lores1)	
			drawblock(&lores1,row,col,value);

		if (!page2)
		{
			if (text || (mixed && (row>=20)))
			{
				XCopyArea(display,text1,w,gc,PIXSIZE*7*col,PIXSIZE*8*row,PIXSIZE*7,PIXSIZE*8,PIXSIZE*7*col,PIXSIZE*8*row);
				if (flush)
					XFlush(display);
				return;
			}
			if (hires) return;
			if (mixed && (row<20))
			{
				XCopyArea(display,lores1,w,gc,PIXSIZE*7*col,PIXSIZE*8*row,PIXSIZE*7,PIXSIZE*8,PIXSIZE*7*col,PIXSIZE*8*row);
				if (flush)
					XFlush(display);
			}
		}
		return;
	}

	if ((adr>=0x800)&&(adr<0xc00))
	{
		if (text2)
			drawchar(&text2,row,col,value);
		if (lores2)	
			drawblock(&lores2,row,col,value);

		if (page2)
		{
			if (text || (mixed && (row>=20)))
			{
				XCopyArea(display,text2,w,gc,PIXSIZE*7*col,PIXSIZE*8*row,PIXSIZE*7,PIXSIZE*8,PIXSIZE*7*col,PIXSIZE*8*row);
				if (flush)
					XFlush(display);
				return;
			}
			if (hires) return;
			if (mixed && (row<20))
			{
				XCopyArea(display,lores2,w,gc,PIXSIZE*7*col,PIXSIZE*8*row,PIXSIZE*7,PIXSIZE*8,PIXSIZE*7*col,PIXSIZE*8*row);
				if (flush)
					XFlush(display);
			}
		}
		return;
	}

	/* hires */
	crow=(adr & 0x1fff)/0x400;

	if ((adr>=0x2000)&&(adr<0x4000))
	{
		if (hires1)
			drawdots(&hires1,row,col,crow,value);

		if ((!text) && (!page2) && (hires))
			if ((!mixed) || (row<20))
			{
				XCopyArea(display,hires1,w,gc,PIXSIZE*7*col,PIXSIZE*(8*row+crow),PIXSIZE*7,PIXSIZE,PIXSIZE*7*col,PIXSIZE*(8*row+crow));
				if (flush)
					XFlush(display);
			}

		return;
	}

	if ((adr>=0x4000)&&(adr<0x6000))
	{
		if (hires2)
			drawdots(&hires2,row,col,crow,value);

		if ((!text) && (page2) && (hires))
			if ((!mixed) || (row<20))
			{
				XCopyArea(display,hires2,w,gc,PIXSIZE*7*col,PIXSIZE*(8*row+crow),PIXSIZE*7,PIXSIZE,PIXSIZE*7*col,PIXSIZE*(8*row+crow));
				if (flush)
					XFlush(display);
			}

		return;
	}
}

void initscreen()
{
	FILE *f;
	int crow,ccol;
	int bits,value;
	int lo,hi;
	int pix1[8],pix2[8];

	/* Open Window */
	displayname=(char *)getenv("DISPLAY");
	display=XOpenDisplay(displayname);
	screen=XDefaultScreen(display);
	rw=XRootWindow(display,screen);
	white=WhitePixel(display,screen);
	black=BlackPixel(display,screen);
	depth=DefaultDepth(display,screen);
	if (depth==1)
		monochrome=1;
	w=XCreateSimpleWindow(display,rw,0,0,PIXSIZE*280,PIXSIZE*256,0,white,black);
	gc=XCreateGC(display,w,0,0);
	XSetForeground(display,gc,white);
	rgc=XCreateGC(display,w,0,0);
	XSetForeground(display,rgc,black);
	XMapWindow(display,w);
	XFlush(display);

	XSelectInput(display,w,KeyPressMask | PointerMotionMask | ButtonPressMask | ButtonReleaseMask | ExposureMask );

	f=fopen("character.rom","r");
	if (f==0)
	{
		printf("can't find 'character.rom'.\n");
		exit(1);
	}
	fread(crom,1,0x0800,f);
	fclose(f);

	/* generate pixmap for charset */
	charset=XCreatePixmap(display,w,PIXSIZE*8*256,PIXSIZE*8,depth);
	XFillRectangle(display,charset,rgc,0,0,PIXSIZE*8*256,PIXSIZE*8);
	for (value=0;value<256;value++)
		for (crow=0;crow<8;crow++)
		{
			bits=crom[8*value+crow] >> 2;
			if (value<128) bits=255-bits;
			for (ccol=0;ccol<8;ccol++)
			{
				if (bits & 0x01)
					XFillRectangle(display,charset,gc,PIXSIZE*(8*value+ccol),PIXSIZE*crow,PIXSIZE,PIXSIZE);
				else
					XFillRectangle(display,charset,rgc,PIXSIZE*(8*value+ccol),PIXSIZE*crow,PIXSIZE,PIXSIZE);
				bits=bits >> 1;
			}
		}

	/* generate pixmap for helpmap */
	helpmap=XCreatePixmap(display,w,PIXSIZE*8*16,PIXSIZE*4,depth);
	if (monochrome)
	{
		/* color 0 (black) */
		XFillRectangle(display,helpmap,rgc,PIXSIZE*8*0,0,PIXSIZE*8,PIXSIZE*4);
		/* color 1 (magenta) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey2_bits,grey2_width,grey2_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*1,0);
		XFreePixmap(display,greymap);
		/* color 2 (darkblue) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey2_bits,grey2_width,grey2_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*2,0);
		XFreePixmap(display,greymap);
		/* color 3 (purple) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey4_bits,grey4_width,grey4_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*3,0);
		XFreePixmap(display,greymap);
		/* color 4 (darkgreen) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey2_bits,grey2_width,grey2_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*4,0);
		XFreePixmap(display,greymap);
		/* color 5 (grey1) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey1_bits,grey1_width,grey1_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*5,0);
		XFreePixmap(display,greymap);
		/* color 6 (mediumblue) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey3_bits,grey3_width,grey3_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*6,0);
		XFreePixmap(display,greymap);
		/* color 7 (lightblue) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey4_bits,grey4_width,grey4_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*7,0);
		XFreePixmap(display,greymap);
		/* color 8 (brown) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey2_bits,grey2_width,grey2_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*8,0);
		XFreePixmap(display,greymap);
		/* color 9 (orange) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey3_bits,grey3_width,grey3_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*9,0);
		XFreePixmap(display,greymap);
		/* color 10 (grey2) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey1_bits,grey1_width,grey1_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*10,0);
		XFreePixmap(display,greymap);
		/* color 11 (pink) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey4_bits,grey4_width,grey4_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*11,0);
		XFreePixmap(display,greymap);
		/* color 12 (lightgreen) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey3_bits,grey3_width,grey3_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*12,0);
		XFreePixmap(display,greymap);
		/* color 13 (yellow) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey4_bits,grey4_width,grey4_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*13,0);
		XFreePixmap(display,greymap);
		/* color 14 (aquamarine) */
		greymap=XCreatePixmapFromBitmapData(display,w,grey4_bits,grey4_width,grey4_height,white,black,depth);
		XCopyArea(display,greymap,helpmap,gc,0,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*14,0);
		XFreePixmap(display,greymap);
		/* color 15 (white) */
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*15,0,PIXSIZE*8,PIXSIZE*4);
	}
	else
	{
		/* allocate colors for lores and hires */
		colormap=XDefaultColormap(display,screen);
		lorescolor[0]=black;
		XSetForeground(display,gc,lorescolor[0]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*0,0,PIXSIZE*8,PIXSIZE*4);
		XAllocNamedColor(display,colormap,"magenta",&color,&exact);
		lorescolor[1]=color.pixel;
		XSetForeground(display,gc,lorescolor[1]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*1,0,PIXSIZE*8,PIXSIZE*4);
		XAllocNamedColor(display,colormap,"MediumBlue",&color,&exact);
		lorescolor[2]=color.pixel;
		XSetForeground(display,gc,lorescolor[2]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*2,0,PIXSIZE*8,PIXSIZE*4);
		XAllocNamedColor(display,colormap,"purple",&color,&exact);
		lorescolor[3]=color.pixel;
		XSetForeground(display,gc,lorescolor[3]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*3,0,PIXSIZE*8,PIXSIZE*4);
		purple=color.pixel;
		XAllocNamedColor(display,colormap,"DarkGreen",&color,&exact);
		lorescolor[4]=color.pixel;
		XSetForeground(display,gc,lorescolor[4]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*4,0,PIXSIZE*8,PIXSIZE*4);
		XAllocNamedColor(display,colormap,"gray50",&color,&exact);
		lorescolor[5]=color.pixel;
		XSetForeground(display,gc,lorescolor[5]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*5,0,PIXSIZE*8,PIXSIZE*4);
		XAllocNamedColor(display,colormap,"DodgerBlue",&color,&exact);
		lorescolor[6]=color.pixel;
		XSetForeground(display,gc,lorescolor[6]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*6,0,PIXSIZE*8,PIXSIZE*4);
		blue=color.pixel;
		XAllocNamedColor(display,colormap,"LightBlue",&color,&exact);
		lorescolor[7]=color.pixel;
		XSetForeground(display,gc,lorescolor[7]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*7,0,PIXSIZE*8,PIXSIZE*4);
		XAllocNamedColor(display,colormap,"brown",&color,&exact);
		lorescolor[8]=color.pixel;
		XSetForeground(display,gc,lorescolor[8]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*8,0,PIXSIZE*8,PIXSIZE*4);
		XAllocNamedColor(display,colormap,"OrangeRed",&color,&exact);
		lorescolor[9]=color.pixel;
		XSetForeground(display,gc,lorescolor[9]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*9,0,PIXSIZE*8,PIXSIZE*4);
		orange=color.pixel;
		XAllocNamedColor(display,colormap,"gray50",&color,&exact);
		lorescolor[10]=color.pixel;
		XSetForeground(display,gc,lorescolor[10]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*10,0,PIXSIZE*8,PIXSIZE*4);
		XAllocNamedColor(display,colormap,"pink",&color,&exact);
		lorescolor[11]=color.pixel;
		XSetForeground(display,gc,lorescolor[11]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*11,0,PIXSIZE*8,PIXSIZE*4);
		XAllocNamedColor(display,colormap,"green",&color,&exact);
		lorescolor[12]=color.pixel;
		XSetForeground(display,gc,lorescolor[12]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*12,0,PIXSIZE*8,PIXSIZE*4);
		green=color.pixel;
		XAllocNamedColor(display,colormap,"yellow",&color,&exact);
		lorescolor[13]=color.pixel;
		XSetForeground(display,gc,lorescolor[13]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*13,0,PIXSIZE*8,PIXSIZE*4);
		XAllocNamedColor(display,colormap,"aquamarine",&color,&exact);
		lorescolor[14]=color.pixel;
		XSetForeground(display,gc,lorescolor[14]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*14,0,PIXSIZE*8,PIXSIZE*4);
		lorescolor[15]=white;
		XSetForeground(display,gc,lorescolor[15]);
		XFillRectangle(display,helpmap,gc,PIXSIZE*8*15,0,PIXSIZE*8,PIXSIZE*4);
	}

	/* generate pixmap for loresset */
	loresset=XCreatePixmap(display,w,PIXSIZE*8*256,PIXSIZE*8,depth);
	for (value=0;value<256;value++)
	{
		hi=value % 16;
		lo=value / 16;

		XCopyArea(display,helpmap,loresset,gc,PIXSIZE*8*hi,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*value,0);
		XCopyArea(display,helpmap,loresset,gc,PIXSIZE*8*lo,0,PIXSIZE*8,PIXSIZE*4,PIXSIZE*8*value,PIXSIZE*4);
	}
	XFreePixmap(display,helpmap);

	/* generate pixmap for hiresset */
	hiresset=XCreatePixmap(display,w,PIXSIZE*8*256,PIXSIZE*2,depth);
	XFillRectangle(display,hiresset,rgc,0,0,PIXSIZE*8*256,PIXSIZE*2);
	for (value=0;value<256;value++)
	{
		if (monochrome)
		{
			bits = value & 0x7f;
			for (ccol=0;ccol<8;ccol++)
			{
				if (bits & 0x01)
					XFillRectangle(display,hiresset,gc,PIXSIZE*(8*value+ccol),0,PIXSIZE,PIXSIZE*2);
				else
					XFillRectangle(display,hiresset,rgc,PIXSIZE*(8*value+ccol),0,PIXSIZE,PIXSIZE*2);
				bits = bits >> 1;
			}
		}
		else
		{
			for (ccol=0;ccol<8;ccol++)
			{
				pix1[ccol]=black;
				pix2[ccol]=black;
			}

			if ((value & 0x80)==0)
			{
				if (value & 0x40) pix1[6]=green;
				if (value & 0x20) pix1[5]=purple;
				if (value & 0x10) pix1[4]=green;
				if (value & 0x08) pix1[3]=purple;
				if (value & 0x04) pix1[2]=green;
				if (value & 0x02) pix1[1]=purple;
				if (value & 0x01) pix1[0]=green;
				if (value & 0x40) pix2[6]=purple;
				if (value & 0x20) pix2[5]=green;
				if (value & 0x10) pix2[4]=purple;
				if (value & 0x08) pix2[3]=green;
				if (value & 0x04) pix2[2]=purple;
				if (value & 0x02) pix2[1]=green;
				if (value & 0x01) pix2[0]=purple;
			}
			else
			{
				if (value & 0x40) pix1[6]=blue;
				if (value & 0x20) pix1[5]=orange;
				if (value & 0x10) pix1[4]=blue;
				if (value & 0x08) pix1[3]=orange;
				if (value & 0x04) pix1[2]=blue;
				if (value & 0x02) pix1[1]=orange;
				if (value & 0x01) pix1[0]=blue;
				if (value & 0x40) pix2[6]=orange;
				if (value & 0x20) pix2[5]=blue;
				if (value & 0x10) pix2[4]=orange;
				if (value & 0x08) pix2[3]=blue;
				if (value & 0x04) pix2[2]=orange;
				if (value & 0x02) pix2[1]=blue;
				if (value & 0x01) pix2[0]=orange;
			}
			
			for (ccol=0;ccol<6;ccol++)
			{
				if ((pix1[ccol]!=black)&&(pix1[ccol+1]!=black))
				{
					pix1[ccol]=white;
					pix1[ccol+1]=white;
				}
				if ((pix2[ccol]!=black)&&(pix2[ccol+1]!=black))
				{
					pix2[ccol]=white;
					pix2[ccol+1]=white;
				}
			}

			if (solid)
			{
				for (ccol=0;ccol<6;ccol++)
				{
					if ((pix1[ccol]==black)&&(pix1[ccol+1]!=white))
						pix1[ccol]=pix1[ccol+1];
					if ((pix2[ccol]==black)&&(pix2[ccol+1]!=white))
						pix2[ccol]=pix2[ccol+1];
				}
				for (ccol=6;ccol>0;ccol--)
				{
					if ((pix1[ccol]==black)&&(pix1[ccol-1]!=white))
						pix1[ccol]=pix1[ccol-1];
					if ((pix2[ccol]==black)&&(pix2[ccol-1]!=white))
						pix2[ccol]=pix2[ccol-1];
				}
			}
			
			for (ccol=0;ccol<8;ccol++)
			{
				XSetForeground(display,gc,pix1[ccol]);
				XFillRectangle(display,hiresset,gc,PIXSIZE*(8*value+ccol),0,PIXSIZE,PIXSIZE);
				XSetForeground(display,gc,pix2[ccol]);
				XFillRectangle(display,hiresset,gc,PIXSIZE*(8*value+ccol),PIXSIZE,PIXSIZE,PIXSIZE);
			}
		}
	}

	XFlush(display);
	XSetForeground(display,gc,white);

	/* Set screens */
	text1=NULL;
	text2=NULL;
	lores1=NULL;
	lores2=NULL;
	hires1=NULL;
	hires2=NULL;

	/* Setup builtin IO */
	keyboard=0;
	speaker=0;

	text=1;
	hires=0;
	page2=0;
	mixed=0;

	redrawscreen();
}
