/* Apple II Simulator Hauptprogramm */
/* Peter Koch, 7.1.93 */

#include "stdio.h"
#include "sys/signal.h"

extern unsigned char accu,xreg,yreg,flags,stack;
extern unsigned short pcnt;

extern unsigned char opcode;
extern unsigned char cpustate;

extern int init6502memory();
extern void init6502();
extern void reset6502();
extern void irq6502();
extern void nmi6502();
extern void exec6502();

int regshow,disasm,beep,keep,solid,monochrome,flush;

quit()
{
	exit(0);
}

main(argc,argv)
int argc;
char **argv;
{
	FILE *f;

	/* Scan arguments */
	disasm=0;
	regshow=0;
	monochrome=0;
	flush=0;
	solid=0;
	keep=0;
	beep=0;
	while (argc>1)
	{
		if (argv[1][0]=='-')
		{
			if (argv[1][1]=='d')
				disasm=1;
			if (argv[1][1]=='r')
				regshow=1;
			if (argv[1][1]=='f')
				flush=1;
			if (argv[1][1]=='m')
				monochrome=1;
			if (argv[1][1]=='k')
				keep=1;
			if (argv[1][1]=='s')
				solid=1;
			if (argv[1][1]=='b')
				beep=1;
			argc--;
			argv++;
		}
		else
		{
			printf("usage: apple2 [-disasm] [-regshow] [-flush] [-mono] [-solid] [-keep] [-beep]\n");
			printf("\n\n");
			printf("       -disasm : disassemble instructions\n");
			printf("       -regshow: show register after each instruction\n");
			printf("       -flush  : flush on every draw instruction\n");
			printf("       -mono   : use black&white graphic on color station\n");
			printf("       -solid  : draw solid colors in hires\n");
			printf("       -keep   : keep invisible screens\n");
			printf("       -beep   : beep when speaker is toggled\n");
			exit(1);
		}
	}

	/* Notice process id */
	f=fopen(".applepid","w");
	fprintf(f,"%d\n",getpid());
	fclose(f);

	/* Init motherboard, screen and slotcards */
	init6502memory();

	/* Establish signal-handlers */
	signal(SIGHUP,reset6502);
	signal(SIGTRAP,irq6502);
	signal(SIGTERM,nmi6502);
	signal(SIGQUIT,quit);

	/* Reset CPU */
	init6502();

	if (disasm)
		cpustate |= 0x10;

	if (regshow)
		cpustate |= 0x08;

	/* loop forever */
	while (1)
		exec6502();
}
